/*
 nRF24 Multi-Network TX
 Version: 0.1
 Author: Alex from insideGadgets (http://www.insidegadgets.com)
 Created: 25/05/2013
 Last Modified: 8/06/2013
 
 */
 
 /*
	Copyright (c) 2007 Stefan Engelke <mbox@stefanengelke.de>

	Permission is hereby granted, free of charge, to any person 
	obtaining a copy of this software and associated documentation 
	files (the "Software"), to deal in the Software without 
	restriction, including without limitation the rights to use, copy, 
	modify, merge, publish, distribute, sublicense, and/or sell copies 
	of the Software, and to permit persons to whom the Software is 
	furnished to do so, subject to the following conditions:
	
	The above copyright notice and this permission notice shall be 
	included in all copies or substantial portions of the Software.
	
	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
	EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
	MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
	NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
	HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
	WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
	DEALINGS IN THE SOFTWARE.
	
	$Id$
*/

#include "mirf.h"
#include "nRF24L01.h"
#include <avr/io.h>
#include <avr/interrupt.h>

// Defines for setting the MiRF registers for transmitting or receiving mode
#define TX_POWERUP mirf_config_register(CONFIG, mirf_CONFIG | ( (1<<PWR_UP) | (0<<PRIM_RX) ) )
#define RX_POWERUP mirf_config_register(CONFIG, mirf_CONFIG | ( (1<<PWR_UP) | (1<<PRIM_RX) ) )
#define POWERDOWN mirf_config_register(CONFIG, mirf_CONFIG | ( (0<<PWR_UP) ) )

extern uint8_t data_in[];
extern uint8_t data_out[];

// Initialise pins
void mirf_init(void) {
	// Define CSN and CE as Output and set them to default
	DDRB |= ((1<<CSN)|(1<<CE));
	mirf_CE_lo;
	mirf_CSN_hi;
}

// Sets the important registers in the MiRF module and powers the module
// in receiving mode
void mirf_config(void) {
	// Set RF channel	
	mirf_config_register(RF_CH, mirf_CH);
	
	// Set length of incoming payload 
	mirf_config_register(RX_PW_P0, mirf_PAYLOAD);
	
	// Disable auto acknowledgement
	mirf_config_register(EN_AA, 0<<ENAA_P0);
	
	// Disable auto retransmit delay
	mirf_config_register(SETUP_RETR, 0);
	
	// Set mirf_ADDR
	mirf_write_register(RX_ADDR_P0, mirf_ADDR, 5);
	mirf_write_register(TX_ADDR, mirf_ADDR, 5);
	
	// Enable RX_ADDR_P0 address matching
	mirf_config_register(EN_RXADDR, (1<<ERX_P0));
	
	// Power up in transmitter mode
	TX_POWERUP;
}

// Flush RX and TX FIFO
void mirf_flush_rx_tx(void) {
	mirf_CSN_lo; // Pull down chip select
	spi_transfer(FLUSH_RX); // Flush RX
	mirf_CSN_hi; // Pull up chip select

	mirf_CSN_lo; // Pull down chip select
	spi_transfer(FLUSH_TX);  // Write cmd to flush tx fifo
	mirf_CSN_hi; // Pull up chip select
}

// Read the status register
uint8_t mirf_status(void) {
	mirf_CSN_lo; // Pull down chip select
	spi_transfer(R_REGISTER | (REGISTER_MASK & STATUS));
	uint8_t status = spi_transfer(NOP); // Read status register
	mirf_CSN_hi; // Pull up chip select
	return status;
}

// Checks if data is available for reading
uint8_t mirf_data_ready(void) {
	mirf_CSN_lo; // Pull down chip select
	spi_transfer(R_REGISTER | (REGISTER_MASK & STATUS));
	uint8_t status = spi_transfer(NOP); // Read status register
	mirf_CSN_hi; // Pull up chip select
	return status & (1<<RX_DR);
}

// Checks if MAX_RT has been reached
uint8_t mirf_max_rt_reached(void) {
	mirf_CSN_lo; // Pull down chip select
	spi_transfer(R_REGISTER | (REGISTER_MASK & STATUS));
	uint8_t status = spi_transfer(NOP); // Read status register
	mirf_CSN_hi; // Pull up chip select
	return status & (1<<MAX_RT);
}

// Write one byte into the MiRF register
void mirf_config_register(uint8_t reg, uint8_t value) {
	mirf_CSN_lo;
	spi_transfer(W_REGISTER | (REGISTER_MASK & reg));
	spi_transfer(value);
	mirf_CSN_hi;
}

// Reads an array of bytes from the MiRF registers.
void mirf_read_register(uint8_t reg, uint8_t *value, uint8_t len) {
	mirf_CSN_lo;
	spi_transfer(R_REGISTER | (REGISTER_MASK & reg));
	spi_read_data(value, len);
	mirf_CSN_hi;
}

// Writes an array of bytes into the MiRF register
void mirf_write_register(uint8_t reg, uint8_t *value, uint8_t len) {
	mirf_CSN_lo;
	spi_transfer(W_REGISTER | (REGISTER_MASK & reg));
	spi_write_data(value, len);
	mirf_CSN_hi;
}

// Send a packet and wait for an ack back, times out after 10 ms
uint8_t send_packet_wait_ack(void) {
	
	uint8_t packet_retries = 0;
	while (packet_retries < 2) {
		
		// Transmit data
		mirf_transmit_data();
		
		// Wait to receive a packet
		if (mirf_receive_data()) {
			// Check the ACK
			if (data_in[FROM] == data_out[TO] && // Coming from who we sent it to
				data_in[TO] == MYADDR && // Packet intended for our address
				data_in[FWD] == NOFWD && // Same FWD type
				data_in[ACK] == ACKRESPOND) { // ACK is respond type
				
				return 1;
			}
		}
		
		packet_retries++;
	}
	
	return 0;
}

// Send data
void mirf_transmit_data(void) {
	TX_POWERUP; // Power up to transmitter mode
	mirf_flush_rx_tx(); // Flush RX and TX FIFO
	
	mirf_CSN_lo;  // Pull down chip select
	spi_transfer(W_TX_PAYLOAD); // Write cmd to write payload
	spi_write_data(data_out, mirf_PAYLOAD); // Write payload
	mirf_CSN_hi; // Pull up chip select
	
	// Reuse TX Payload, it takes 5ms for the receiver to read out RX Payload so re-send the packet a few times to ensure they get it
	for (int x = 0; x < 8; x++) {
		mirf_CE_hi; // Start transmission
		_delay_us(15);
		mirf_CE_lo;
		
		_delay_us(500); // Wait until packet is transmitted
		
		mirf_CSN_lo;  // Pull down chip select
		spi_transfer(REUSE_TX_PL); // Write cmd to reuse last payload
		mirf_CSN_hi; // Pull up chip select
	}
	
	// We always flush TX/RX before sending/receiving so we don't need to do it here to void the Reuse TX PL
}

// Receive data
uint8_t mirf_receive_data(void) {
	RX_POWERUP; // Power up to receiver mode
	mirf_flush_rx_tx(); // Flush RX and TX FIFO
	mirf_CE_hi; // Start listening
	
	for (int x = 0; x < 32; x++) {
		data_in[x] = 0; // Clear the data_in
	}
	
	byte waitcount = 0;
	while (!(mirf_status() & (1<<RX_DR))) { // Wait until data arrives
		_delay_us(50);
		waitcount++;
		if (waitcount >= 50) { // If we haven't heard back in 5ms, exit
			mirf_CE_lo; // Stop listening
			return 0;
		}
	}
	mirf_CE_lo; // Stop listening
	
	mirf_CSN_lo; // Pull down chip select
	spi_transfer(R_RX_PAYLOAD); // Send cmd to read rx payload
	spi_read_data(data_in, mirf_PAYLOAD); // Read payload
	mirf_CSN_hi; // Pull up chip select
	mirf_config_register(STATUS,(1<<RX_DR)); // Reset status register
	
	return 1;
}
